# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import sys

import numpy as np
import PIL
import pytest
from common_vision_utils import (FRIDGE_MULTILABEL_TARGETS, ImageTransformEnum,
                                 ImageTransforms, create_dummy_model,
                                 create_image_classification_pipeline,
                                 create_pytorch_vision_model,
                                 gridify_fridge_multilabel_labels,
                                 load_flowers_dataset, load_fridge_dataset,
                                 load_fridge_object_detection_dataset,
                                 load_imagenet_dataset, load_imagenet_labels,
                                 load_mnist_dataset,
                                 load_multilabel_fridge_dataset,
                                 retrieve_fridge_object_detection_model,
                                 retrieve_or_train_fridge_model)
from ml_wrappers import wrap_model
from ml_wrappers.common.constants import Device
from rai_vision_insights_validator import validate_rai_vision_insights

from responsibleai.feature_metadata import FeatureMetadata
from responsibleai_vision import ModelTask, RAIVisionInsights
from responsibleai_vision.common.constants import (ExplainabilityDefaults,
                                                   ImageColumns, ImageModes)

DEFAULT_MAX_EVALS = ExplainabilityDefaults.DEFAULT_MAX_EVALS
DEFAULT_NUM_MASKS = ExplainabilityDefaults.DEFAULT_NUM_MASKS
DEFAULT_MASK_RES = ExplainabilityDefaults.DEFAULT_MASK_RES


class TestRAIVisionInsights(object):

    def test_rai_insights_image_classification_imagenet(self):
        data = load_imagenet_dataset()
        pred = create_image_classification_pipeline()
        task_type = ModelTask.IMAGE_CLASSIFICATION
        class_names = load_imagenet_labels()
        run_rai_insights(pred, data[:3], ImageColumns.LABEL,
                         task_type, class_names, image_mode=ImageModes.RGB)

    @pytest.mark.parametrize('max_evals', [None, 10, 200])
    def test_rai_insights_image_classification_max_evals(self, max_evals):
        data = load_imagenet_dataset()
        pred = create_image_classification_pipeline()
        task_type = ModelTask.IMAGE_CLASSIFICATION
        class_names = load_imagenet_labels()
        # run on a single image to avoid running out of memory on
        # test machines
        run_rai_insights(pred, data[:1], ImageColumns.LABEL,
                         task_type, class_names, image_mode=ImageModes.RGB,
                         test_explainer=True, max_evals=max_evals)

    @pytest.mark.parametrize('max_evals', [-100, -1, 0])
    def test_rai_insights_invalid_max_evals(self, max_evals):
        data = load_imagenet_dataset()
        pred = create_image_classification_pipeline()
        task_type = ModelTask.IMAGE_CLASSIFICATION
        class_names = load_imagenet_labels()
        with pytest.raises(ValueError,
                           match="max_evals must be greater than 0"):
            run_rai_insights(pred, data[:1], ImageColumns.LABEL,
                             task_type, class_names,
                             image_mode=ImageModes.RGB,
                             test_explainer=True, max_evals=max_evals)

    def test_rai_insights_image_classification_fridge(self):
        data = load_fridge_dataset()
        try:
            model = retrieve_or_train_fridge_model(data)
        except Exception as e:
            print("Failed to retrieve or load Fastai model, force training")
            print("Inner exception message on retrieving model: {}".format(e))
            model = retrieve_or_train_fridge_model(data, force_train=True)
        task_type = ModelTask.IMAGE_CLASSIFICATION
        class_names = data[ImageColumns.LABEL.value].unique()
        run_rai_insights(model, data[:3], ImageColumns.LABEL,
                         task_type, class_names, test_error_analysis=True)

    def test_rai_insights_image_classification_mnist(self):
        train_data, test_data = load_mnist_dataset()
        model = create_pytorch_vision_model(train_data, test_data)
        task_type = ModelTask.IMAGE_CLASSIFICATION
        class_names = train_data[ImageColumns.LABEL.value].unique()
        run_rai_insights(
            model, test_data[:3], ImageColumns.LABEL,
            task_type, class_names)

    def test_rai_insights_multilabel_image_classification_fridge(self):
        data = load_multilabel_fridge_dataset()
        try:
            model = retrieve_or_train_fridge_model(data, multilabel=True)
        except Exception as e:
            print("Failed to retrieve or load Fastai model, force training")
            print("Inner exception message on retrieving model: {}".format(e))
            model = retrieve_or_train_fridge_model(
                data, force_train=True, multilabel=True)
        data = gridify_fridge_multilabel_labels(data)
        task_type = ModelTask.MULTILABEL_IMAGE_CLASSIFICATION
        run_rai_insights(model, data[:3], FRIDGE_MULTILABEL_TARGETS,
                         task_type, test_error_analysis=True)

    @pytest.mark.skipif(
        sys.version_info[:2] == (3, 7) or sys.version_info[:2] == (3, 8),
        reason='Incompatible torch state update for fridge model')
    @pytest.mark.parametrize('num_masks', [None, 25, DEFAULT_NUM_MASKS])
    @pytest.mark.parametrize('mask_res', [None, DEFAULT_MASK_RES, 8])
    def test_rai_insights_object_detection_fridge(self, num_masks, mask_res):
        data = load_fridge_object_detection_dataset()
        model = retrieve_fridge_object_detection_model(
            load_fridge_weights=True
        )
        task_type = ModelTask.OBJECT_DETECTION
        class_names = np.array(['can', 'carton',
                                'milk_bottle', 'water_bottle'])
        run_rai_insights(model, data[:2], ImageColumns.LABEL,
                         task_type, class_names,
                         num_masks=num_masks, mask_res=mask_res,
                         test_error_analysis=True)

    def test_rai_insights_object_detection_jagged_list(self):
        data = load_fridge_object_detection_dataset()
        model = retrieve_fridge_object_detection_model(
            load_fridge_weights=True
        )
        task_type = ModelTask.OBJECT_DETECTION
        wrapped_model = wrap_model(model, data, ModelTask.OBJECT_DETECTION)

        class DummyPredictWrapper(object):
            def __init__(self, model):
                self.model = model
                self._model = model._model

            def to(self, dummy):
                pass

            def predict(self, X):
                prediction = self.model.predict(X).tolist()
                if len(prediction) == 1:
                    # fix ndim error for some versions of pandas in tests
                    prediction = prediction[0]
                return prediction

            def predict_proba(self, X):
                return self.model.predict_proba(X)

        wrapped_model = DummyPredictWrapper(wrapped_model)
        class_names = np.array(['can', 'carton',
                                'milk_bottle', 'water_bottle'])
        # test case where there are different numbers of objects in labels
        images = ['10.jpg', '29.jpg', '92.jpg']
        images = ["./data/odFridgeObjects/images/{}".format(i) for i in images]
        data = data.loc[data[ImageColumns.IMAGE.value].isin(images)]
        data = data.reset_index(drop=True)
        run_rai_insights(wrapped_model, data, ImageColumns.LABEL,
                         task_type, class_names,
                         test_error_analysis=True,
                         ignore_index=True)

    @pytest.mark.parametrize('num_masks', [-100, -1, 0])
    def test_rai_insights_invalid_num_masks(self, num_masks):
        data = load_fridge_object_detection_dataset()
        model = retrieve_fridge_object_detection_model()
        task_type = ModelTask.OBJECT_DETECTION
        class_names = np.array(['can', 'carton',
                                'milk_bottle', 'water_bottle'])
        with pytest.raises(ValueError,
                           match="num_masks must be greater than 0"):
            run_rai_insights(model, data[:1], ImageColumns.LABEL,
                             task_type, class_names, num_masks=num_masks)

    @pytest.mark.parametrize('mask_res', [-100, -1, 0])
    def test_rai_insights_invalid_mask_res(self, mask_res):
        data = load_fridge_object_detection_dataset()
        model = retrieve_fridge_object_detection_model()
        task_type = ModelTask.OBJECT_DETECTION
        class_names = np.array(['can', 'carton',
                                'milk_bottle', 'water_bottle'])
        with pytest.raises(ValueError,
                           match="mask_res must be greater than 0"):
            run_rai_insights(model, data[:1], ImageColumns.LABEL,
                             task_type, class_names, mask_res=mask_res)

    @pytest.mark.parametrize('device', [None, 'cpu'])
    def test_rai_insights_device(self, device):
        data = load_fridge_object_detection_dataset()
        model = retrieve_fridge_object_detection_model()
        task_type = ModelTask.OBJECT_DETECTION
        class_names = np.array(['can', 'carton',
                                'milk_bottle', 'water_bottle'])
        run_rai_insights(model, data[:1], ImageColumns.LABEL,
                         task_type, class_names, device=device)

    @pytest.mark.parametrize('device', ["bad_device", ""])
    def test_rai_insights_invalid_device(self, device):
        data = load_fridge_object_detection_dataset()
        model = retrieve_fridge_object_detection_model()
        task_type = ModelTask.OBJECT_DETECTION
        class_names = np.array(['can', 'carton',
                                'milk_bottle', 'water_bottle'])
        with pytest.raises(ValueError, match="Selected device is invalid"):
            run_rai_insights(model, data[:1], ImageColumns.LABEL,
                             task_type, class_names, device=device)

    @pytest.mark.skip("This test fails in the build due to \
                      incompatibility between fastai and pytorch \
                      2.0.0. TODO: fix may be to ping pytorch <2.0.0 \
                      in the build until fastai updates.")
    def test_rai_insights_object_detection_fridge_label_format(self):
        data = load_fridge_object_detection_dataset()
        model = retrieve_fridge_object_detection_model()
        task_type = ModelTask.OBJECT_DETECTION
        class_names = np.array(['can', 'carton',
                                'milk_bottle', 'water_bottle'])

        rai_insights = RAIVisionInsights(model, data[:3],
                                         ImageColumns.LABEL,
                                         task_type=task_type,
                                         classes=class_names)
        y = [
            [
                [1, 100, 200, 300, 400, 0.95],
                [2, 100, 200, 300, 400, 0.95],
                [1, 100, 200, 300, 400, 0.95]
            ],
            [
                [1, 100, 200, 300, 400, 0.95],
                [2, 100, 200, 300, 400, 0.95],
            ]
        ]
        result = [
            [2, 1, 0, 0],
            [1, 1, 0, 0]
        ]
        assert rai_insights._format_od_labels(y, class_names) == result

    @pytest.mark.skipif(sys.platform == 'darwin',
                        reason='torch version downgrade on macos')
    @pytest.mark.parametrize("path, transform, size", [
        ("./data/odFridgeObjects/img_transforms_large",
         ImageTransformEnum.RESIZE,
         (1000, 1000)),
        ("./data/odFridgeObjects/img_transforms_gray",
         ImageTransformEnum.GRAYSCALE,
         None),
        ("./data/odFridgeObjects/img_transforms_opacity",
         ImageTransformEnum.OPACITY,
         None),
        ("./data/odFridgeObjects/img_transforms_blackout",
         ImageTransformEnum.BLACKOUT,
         None),
        ("./data/odFridgeObjects/img_transforms_png",
         ImageTransformEnum.PNG,
         None),
    ])
    def test_rai_insights_object_detection_fridge_image_transforms(self,
                                                                   path,
                                                                   transform,
                                                                   size):
        data = load_fridge_object_detection_dataset()[:10]
        data = ImageTransforms(data).apply_transformation(path,
                                                          transform,
                                                          size)
        model = retrieve_fridge_object_detection_model()
        task_type = ModelTask.OBJECT_DETECTION
        class_names = np.array(['can', 'carton',
                                'milk_bottle', 'water_bottle'])
        dropped_cols_num = [i for i in range(0, 10)]
        dropped_features = ["{}".format(i) for i in dropped_cols_num]
        # rename column names to strings since RAI validation fails otherwise
        data = data.rename(columns={i: j for i, j in zip(
            dropped_cols_num, dropped_features)}).reset_index(drop=True)
        run_rai_insights(model, data[:3], ImageColumns.LABEL,
                         task_type, class_names,
                         dropped_features=dropped_features)

    @pytest.mark.parametrize(
        'upscale',
        [
            pytest.param(
                True,
                marks=pytest.mark.skip(
                    'Insufficient memory on test machines to load images')),
            False
        ])
    def test_jagged_image_sizes(self, upscale):
        if upscale:
            PIL.Image.MAX_IMAGE_PIXELS = None
        data = load_flowers_dataset(upscale=upscale)
        model = create_dummy_model(data)
        test_data = data
        class_names = data[ImageColumns.LABEL.value].unique()
        task_type = ModelTask.IMAGE_CLASSIFICATION
        run_rai_insights(model, test_data, ImageColumns.LABEL,
                         task_type, class_names, upscale=upscale)


def run_rai_insights(model, test_data, target_column,
                     task_type, classes=None, test_explainer=False,
                     test_error_analysis=False,
                     image_mode=None, dropped_features=None,
                     upscale=False, max_evals=DEFAULT_MAX_EVALS,
                     num_masks=DEFAULT_NUM_MASKS,
                     mask_res=DEFAULT_MASK_RES,
                     device=Device.AUTO.value,
                     ignore_index=False):
    feature_metadata = None
    if dropped_features:
        feature_metadata = FeatureMetadata(dropped_features=dropped_features)
    image_width = None
    if upscale:
        image_width = 2
    rai_insights = RAIVisionInsights(model, test_data,
                                     target_column,
                                     task_type=task_type,
                                     classes=classes,
                                     image_mode=image_mode,
                                     feature_metadata=feature_metadata,
                                     image_width=image_width,
                                     max_evals=max_evals,
                                     num_masks=num_masks,
                                     mask_res=mask_res,
                                     device=device)
    # Note: this seems too resource-intensive
    # TODO: re-add when we get beefier test machines
    if test_explainer:
        rai_insights.explainer.add()
    if test_error_analysis:
        rai_insights.error_analysis.add()
    if test_explainer or test_error_analysis:
        rai_insights.compute()
    rai_insights.get_data()
    # Validate
    validate_rai_vision_insights(
        rai_insights, test_data,
        target_column, task_type,
        ignore_index)
    if task_type == ModelTask.OBJECT_DETECTION:
        selection_indexes = [[0]]
        class_name = classes[0]
        iou_threshold = 70
        object_detection_cache = {}
        aggregate_methods = ['macro', 'micro']
        for aggregate_method in aggregate_methods:
            metrics = rai_insights.compute_object_detection_metrics(
                selection_indexes, aggregate_method, class_name, iou_threshold,
                object_detection_cache)
            assert len(metrics) == 2



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

import shutil
from pathlib import Path
from tempfile import TemporaryDirectory
from unittest.mock import patch

import numpy as np
import PIL
import pytest
from common_vision_utils import (DummyFlowersPipelineSerializer,
                                 ImageClassificationPipelineSerializer,
                                 ObjectDetectionPipelineSerializer,
                                 create_dummy_model,
                                 create_image_classification_pipeline,
                                 load_flowers_dataset,
                                 load_fridge_object_detection_dataset,
                                 load_imagenet_dataset, load_imagenet_labels,
                                 retrieve_fridge_object_detection_model)
from rai_vision_insights_validator import run_and_validate_serialization

from responsibleai_vision import ModelTask, RAIVisionInsights
from responsibleai_vision.common.constants import ImageColumns

FRIDGE_CLASS_NAMES = np.array(['can', 'carton',
                               'milk_bottle', 'water_bottle'])


class FakeImageDownloader:
    def __init__(self, test_mltable_path):
        self._images_df = self.get_data()

    def get_data(self):
        return None


class TestRAIVisionInsightsSaveAndLoadScenarios(object):

    def test_rai_insights_empty_save_load_save(self):
        data = load_imagenet_dataset()
        pred = create_image_classification_pipeline()
        task_type = ModelTask.IMAGE_CLASSIFICATION
        class_names = load_imagenet_labels()
        test = data[:3]
        label = ImageColumns.LABEL
        serializer = ImageClassificationPipelineSerializer()

        run_and_validate_serialization(
            pred, test, task_type, class_names, label, serializer)

    @pytest.mark.skip("Insufficient memory on test machines to load images")
    def test_rai_insights_large_images_save_load_save(self):
        PIL.Image.MAX_IMAGE_PIXELS = None
        data = load_flowers_dataset(upscale=True)
        model = create_dummy_model(data)
        test = data
        class_names = data[ImageColumns.LABEL.value].unique()
        task_type = ModelTask.IMAGE_CLASSIFICATION
        label = ImageColumns.LABEL
        serializer = DummyFlowersPipelineSerializer()

        image_width = 2
        run_and_validate_serialization(
            model, test, task_type, class_names, label, serializer,
            image_width)

    def test_loading_rai_insights_without_model_file(self):
        data = load_imagenet_dataset()
        pred = create_image_classification_pipeline()
        task_type = ModelTask.IMAGE_CLASSIFICATION
        class_names = load_imagenet_labels()
        test = data[:3]
        label = ImageColumns.LABEL
        serializer = ImageClassificationPipelineSerializer()

        rai_insights = RAIVisionInsights(
            pred, test, label,
            task_type=task_type,
            classes=class_names,
            serializer=serializer)

        with TemporaryDirectory() as tmpdir:
            assert rai_insights.model is not None
            save_path = Path(tmpdir) / "rai_insights"
            rai_insights.save(save_path)

            # Remove the model.pkl file to cause an exception to occur
            # while loading the model.
            model_name = 'image-classification-model'
            model_pkl_path = Path(tmpdir) / "rai_insights" / model_name
            shutil.rmtree(model_pkl_path)
            match_msg = 'No file or directory found'
            with pytest.raises(OSError, match=match_msg):
                without_model_rai_insights = RAIVisionInsights.load(save_path)
                assert without_model_rai_insights.model is None

    @pytest.mark.parametrize('automl_format', [True, False])
    def test_rai_insights_object_detection(self, automl_format):
        data = load_fridge_object_detection_dataset(automl_format)
        model = retrieve_fridge_object_detection_model(
            load_fridge_weights=True
        )
        task_type = ModelTask.OBJECT_DETECTION
        test = data[:3]
        label = ImageColumns.LABEL
        serializer = ObjectDetectionPipelineSerializer()

        run_and_validate_serialization(
            model, test, task_type, FRIDGE_CLASS_NAMES, label,
            serializer, ignore_test_data=True)

    def test_rai_insights_image_downloader_object_detection(self):
        data = load_fridge_object_detection_dataset(True)
        model = retrieve_fridge_object_detection_model(
            load_fridge_weights=True
        )
        task_type = ModelTask.OBJECT_DETECTION
        test = data[:3]
        label = ImageColumns.LABEL
        serializer = ObjectDetectionPipelineSerializer()

        get_data = ('test_rai_vision_insights_save_and_load_scenarios'
                    '.FakeImageDownloader.get_data')
        with patch(get_data) as mock_images_df:
            mock_images_df.return_value = test.copy()
            run_and_validate_serialization(
                model, test, task_type, FRIDGE_CLASS_NAMES, label,
                serializer, ignore_test_data=True,
                image_downloader=FakeImageDownloader)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from common_vision_utils import (load_flowers_dataset, load_fridge_dataset,
                                 load_fridge_object_detection_dataset,
                                 load_imagenet_dataset)

from responsibleai_vision.common.constants import (ExtractedFeatures,
                                                   ImageColumns, ImageModes)
from responsibleai_vision.utils.feature_extractors import extract_features

MEAN_PIXEL_VALUE = ExtractedFeatures.MEAN_PIXEL_VALUE.value
FRIDGE_METADATA_FEATURES = [
    'Make', 'ResolutionUnit', 'ImageLength', 'ExifOffset', 'Model',
    'GPSInfo', 'ImageWidth', 'DateTime', 'YCbCrPositioning',
    'Software', 'Orientation'
]


def validate_extracted_features(extracted_features, feature_names,
                                expected_feature_names, data):
    assert len(extracted_features) == len(data)
    assert feature_names[0] == expected_feature_names[0]
    for i in range(1, len(feature_names)):
        assert feature_names[i] in expected_feature_names
    assert len(feature_names) == len(expected_feature_names)
    assert len(extracted_features[0]) == len(feature_names)


def extract_dataset_features(data):
    return extract_features(data, ImageColumns.LABEL, ImageModes.RGB, None)


class TestFeatureExtractors(object):
    def test_extract_features_fridge_object_detection(self):
        data = load_fridge_object_detection_dataset(automl_format=False)
        extracted_features, feature_names = extract_dataset_features(data)
        expected_feature_names = [MEAN_PIXEL_VALUE] + FRIDGE_METADATA_FEATURES
        validate_extracted_features(extracted_features, feature_names,
                                    expected_feature_names, data)

    def test_extract_features_fridge_metadata(self):
        data = load_fridge_dataset()
        extracted_features, feature_names = extract_dataset_features(data)
        expected_feature_names = [MEAN_PIXEL_VALUE] + FRIDGE_METADATA_FEATURES
        validate_extracted_features(extracted_features, feature_names,
                                    expected_feature_names, data)

    def test_extract_features_imagenet_metadata(self):
        data = load_imagenet_dataset()
        extracted_features, feature_names = extract_dataset_features(data)
        expected_feature_names = [MEAN_PIXEL_VALUE]
        validate_extracted_features(extracted_features, feature_names,
                                    expected_feature_names, data)

    def test_extract_features_flowers_metadata(self):
        data = load_flowers_dataset(upscale=False)
        extracted_features, feature_names = extract_dataset_features(data)
        expected_feature_names = [MEAN_PIXEL_VALUE]
        validate_extracted_features(extracted_features, feature_names,
                                    expected_feature_names, data)

    def test_extract_features_mixed_exif_XPComment_metadata(self):
        data = load_fridge_dataset(add_extra_mixed_metadata=True)
        extracted_features, feature_names = extract_dataset_features(data)
        expected_feature_names = [MEAN_PIXEL_VALUE, 'XPComment']
        expected_feature_names += FRIDGE_METADATA_FEATURES
        validate_extracted_features(extracted_features, feature_names,
                                    expected_feature_names, data)



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

from collections import Counter
from http.client import HTTPMessage
from math import isclose
from unittest.mock import Mock, patch
from urllib.parse import urlparse

import numpy as np
from common_vision_utils import load_fridge_object_detection_dataset

from responsibleai_vision.common.constants import ImageColumns
from responsibleai_vision.utils.image_reader import \
    _get_retry_session as image_reader_get_retry_session
from responsibleai_vision.utils.image_reader import \
    _requests_sessions as image_reader_requests_sessions
from responsibleai_vision.utils.image_reader import get_all_exif_feature_names
from responsibleai_vision.utils.image_utils import (
    BOTTOM_X, BOTTOM_Y, HEIGHT, IS_CROWD, TOP_X, TOP_Y, WIDTH, classes_to_dict,
    transform_object_detection_labels)

LABEL = ImageColumns.LABEL.value
IMAGE_DETAILS = ImageColumns.IMAGE_DETAILS.value
TOL = 1


class TestImageUtils(object):
    def test_transform_object_detection_labels(self):
        data = load_fridge_object_detection_dataset(automl_format=True)
        class_names = np.array(
            ["can", "carton", "milk_bottle", "water_bottle"]
        )
        data_transformed = transform_object_detection_labels(
            data.copy(), LABEL, class_names
        )
        label_dict = classes_to_dict(class_names)
        for i in range(len(data)):
            original_label = data[LABEL][i]
            label = data_transformed[LABEL][i]
            image_details = data[IMAGE_DETAILS][i]
            width = image_details[WIDTH]
            height = image_details[HEIGHT]
            assert isinstance(label, list)
            for j in range(len(label)):
                label_j = label[j]
                assert len(label_j) == 6
                o_label_j = original_label[j]
                assert label_j[0] == label_dict[o_label_j[LABEL]]
                # use is close for off by one errors due to rounding
                assert isclose(
                    label_j[1], o_label_j[TOP_X] * width, abs_tol=TOL
                )
                assert isclose(
                    label_j[2], o_label_j[TOP_Y] * height, abs_tol=TOL
                )
                assert isclose(
                    label_j[3], o_label_j[BOTTOM_X] * width, abs_tol=TOL
                )
                assert isclose(
                    label_j[4], o_label_j[BOTTOM_Y] * height, abs_tol=TOL
                )
                assert label_j[5] == o_label_j[IS_CROWD]

    def test_retry_sessions_match_domain_count(self):
        sessions_before_test = len(image_reader_requests_sessions)
        urls = [f"https://{i}.com/image.png" for i in range(10)]
        duplicates = urls.copy()
        urls.extend(duplicates)
        domains = [urlparse(d.lower()).netloc for d in urls]
        domain_unique_count = len(Counter(domains))

        for url in urls:
            image_reader_get_retry_session(url)

        new_session_count = len(image_reader_requests_sessions)
        new_session_count -= sessions_before_test
        assert new_session_count == domain_unique_count

    @patch("urllib3.connectionpool.HTTPConnectionPool._make_request")
    def test_retry_sessions_retries_on_conn_failure(self, request_mock):
        url = "http://mock.com/test.png"

        # requests_send_mock.side_effect =
        request_mock.side_effect = [
            IOError(),
            Mock(status=500, msg=HTTPMessage("test")),
            Mock(status=200, msg=HTTPMessage("test")),
        ]
        session = image_reader_get_retry_session(url)

        assert session.get(url).status_code == 200

    def test_get_all_exif_feature_names(self):
        image_dataset = load_fridge_object_detection_dataset().head(2)
        exif_feature_names = get_all_exif_feature_names(image_dataset)
        assert len(exif_feature_names) == 11
        assert set(exif_feature_names) == \
            set(['Orientation', 'ExifOffset', 'ImageWidth', 'GPSInfo',
                 'Model', 'DateTime', 'YCbCrPositioning', 'ImageLength',
                 'ResolutionUnit', 'Software', 'Make'])



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Utilities for reading images."""

import base64
from io import BytesIO
from typing import Any, Tuple, Union
from urllib.parse import urlparse

import requests
from numpy import asarray
from PIL import Image
from PIL.ExifTags import TAGS
from requests.adapters import HTTPAdapter, Retry

from responsibleai_vision.common.constants import (AutoMLImagesModelIdentifier,
                                                   CommonTags)

# domain mapped session for reuse
_requests_sessions = {}


def _get_retry_session(url):
    domain = urlparse(url.lower()).netloc
    if domain in _requests_sessions:
        return _requests_sessions[domain]

    session = requests.Session()
    retries = Retry(
        total=5, backoff_factor=1, status_forcelist=[500, 502, 503, 504]
    )
    session.mount("http://", HTTPAdapter(max_retries=retries))
    session.mount("https://", HTTPAdapter(max_retries=retries))
    _requests_sessions[domain] = session

    return session


def get_image_pointer_from_path(image_path):
    """Get image pointer from path.

    :param image_path: The path to the image.
    :type image_path: str
    :return: The image open pointer.
    :rtype: str or bytes
    """
    image_open_pointer = image_path
    if image_path.startswith("http://") or image_path.startswith("https://"):
        response = _get_retry_session(image_path).get(image_path)
        image_open_pointer = BytesIO(response.content)
    return image_open_pointer


def get_image_from_path(image_path, image_mode):
    """Get image from path.

    :param image_path: The path to the image.
    :type image_path: str
    :param image_mode: The mode to open the image in.
        See pillow documentation for all modes:
        https://pillow.readthedocs.io/en/stable/handbook/concepts.html
    :type image_mode: str
    :return: The image as a numpy array.
    :rtype: numpy.ndarray
    """
    image_open_pointer = get_image_pointer_from_path(image_path)
    with Image.open(image_open_pointer) as im:
        if image_mode is not None:
            im = im.convert(image_mode)
        image_array = asarray(im)
    return image_array


def get_all_exif_feature_names(image_dataset):
    """Get all exif feature names.

    :param image_dataset: The image dataset.
    :type image_dataset: pandas.DataFrame
    :return: The list of exif feature names.
    :rtype: List[Union[str, int]]]
    """
    exif_feature_names = set()
    for i in range(image_dataset.shape[0]):
        image = image_dataset.iloc[i, 0]
        if isinstance(image, str):
            image_pointer_path = get_image_pointer_from_path(image)
            with Image.open(image_pointer_path) as im:
                exifdata = im.getexif()
                for tag_id in exifdata:
                    # get the tag name, instead of human unreadable tag id
                    tag = TAGS.get(tag_id, tag_id)
                    if tag not in image_dataset.columns:
                        data = exifdata.get(tag_id)
                        if isinstance(data, str) or \
                           isinstance(data, int) or \
                           isinstance(data, float) or \
                           isinstance(data, bytes):
                            exif_feature_names.add(tag)
    return list(exif_feature_names)


def get_base64_string_from_path(
    img_path: str, return_image_size: bool = False
) -> Union[str, Tuple[str, Tuple[int, int]]]:
    """Load and convert pillow image to base64-encoded image

    :param img_path: image path
    :type img_path: str
    :param return_image_size: true if image size should also be returned
    :type return_image_size: bool
    :return: base64-encoded image OR base64-encoded image and image size
    :rtype: Union[str, Tuple[str, Tuple[int, int]]]
    """
    try:
        img = Image.open(img_path)
    except Exception as e:
        print("file not found", str(e))
        import urllib.request

        urllib.request.urlretrieve(img_path, "tempfile")
        img = Image.open("tempfile")
    imgio = BytesIO()
    img.save(imgio, img.format)
    img_str = base64.b64encode(imgio.getvalue())
    if return_image_size:
        return img_str.decode(CommonTags.IMAGE_DECODE_UTF_FORMAT), img.size
    return img_str.decode(CommonTags.IMAGE_DECODE_UTF_FORMAT)


def is_automl_image_model(model: Any) -> bool:
    """Check whether the model is automl images mlflow type

    :param model: Model object
    :type model: supported model types
    :return: True if automl model type else False
    :rtype: bool
    """
    automl_image_model = False
    model_type = str(type(model))
    if model_type.endswith(
        AutoMLImagesModelIdentifier.AUTOML_IMAGE_CLASSIFICATION_MODEL
    ) or model_type.endswith(
        AutoMLImagesModelIdentifier.AUTOML_OBJECT_DETECTION_MODEL
    ):
        automl_image_model = True
    return automl_image_model



# Copyright (c) Microsoft Corporation
# Licensed under the MIT License.

"""Defines the feature extractors."""

import warnings
from typing import Optional

import pandas as pd
from PIL import Image
from PIL.ExifTags import TAGS
from tqdm import tqdm

from responsibleai.feature_metadata import FeatureMetadata
from responsibleai_vision.common.constants import ExtractedFeatures
from responsibleai_vision.utils.image_reader import (
    get_all_exif_feature_names, get_image_from_path,
    get_image_pointer_from_path)

MEAN_PIXEL_VALUE = ExtractedFeatures.MEAN_PIXEL_VALUE.value
MAX_CUSTOM_LEN = 100


def extract_features(image_dataset: pd.DataFrame,
                     target_column: str,
                     image_mode: str = None,
                     feature_metadata: Optional[FeatureMetadata] = None):
    '''Extract tabular data features from the image dataset.

    :param image_dataset: A pandas dataframe containing the image data.
    :type image_dataset: pandas.DataFrame
    :param target_column: The name of the label column or list of columns.
        This is a list of columns for multilabel models.
    :type target_column: str or list[str]
    :param image_mode: The mode to open the image in.
        See pillow documentation for all modes:
        https://pillow.readthedocs.io/en/stable/handbook/concepts.html
    :type image_mode: str
    :param feature_metadata: Feature metadata for the dataset
        to identify different kinds of features.
    :type feature_metadata: Optional[FeatureMetadata]
    :return: The list of extracted features and the feature names.
    :rtype: list, list
    '''
    results = []
    dropped_features = feature_metadata.dropped_features \
        if feature_metadata else None
    if feature_metadata and feature_metadata.categorical_features is None:
        feature_metadata.categorical_features = []
    exif_feature_names = get_all_exif_feature_names(image_dataset)
    feature_names = [MEAN_PIXEL_VALUE] + exif_feature_names

    # append all feature names other than target column and label
    column_names = image_dataset.columns
    has_dropped_features = dropped_features is not None
    start_meta_index = 2
    if isinstance(target_column, list):
        start_meta_index = len(target_column) + 1
    for j in range(start_meta_index, image_dataset.shape[1]):
        if has_dropped_features and column_names[j] in dropped_features:
            continue
        feature_names.append(column_names[j])

    blacklisted_tags = {}
    # append all features
    for i in tqdm(range(image_dataset.shape[0])):
        image = image_dataset.iloc[i, 0]
        if isinstance(image, str):
            image_arr = get_image_from_path(image, image_mode)
            mean_pixel_value = image_arr.mean()
        else:
            mean_pixel_value = image.mean()
        row_feature_values = [mean_pixel_value] + \
            [None] * len(exif_feature_names)

        # append all exif features
        if isinstance(image, str):
            image_pointer_path = get_image_pointer_from_path(image)
            with Image.open(image_pointer_path) as im:
                exifdata = im.getexif()
                for tag_id in exifdata:
                    # get the tag name, instead of human unreadable tag id
                    tag = TAGS.get(tag_id, tag_id)
                    data = exifdata.get(tag_id)
                    # decode bytes
                    if isinstance(data, bytes):
                        data = data.decode()
                        if len(data) > MAX_CUSTOM_LEN:
                            data = data[:MAX_CUSTOM_LEN] + '...'
                    if isinstance(data, str):
                        if not feature_metadata:
                            feature_metadata = FeatureMetadata()
                            feature_metadata.categorical_features = []
                        if tag in feature_names:
                            feature_metadata.categorical_features.append(
                                str(tag))
                            tag_index = feature_names.index(tag)
                            row_feature_values[tag_index] = data
                        else:
                            # in theory this should now never happen with
                            # latest code, but adding this check for safety
                            if tag not in blacklisted_tags:
                                blacklisted_tags.add(tag)
                                warnings.warn(
                                    f'Exif tag {tag} could not be found '
                                    'in the feature names. Ignoring tag '
                                    'from extracted metadata.')
                    elif isinstance(data, int) or isinstance(data, float):
                        row_feature_values[feature_names.index(tag)] = data

        # append all features other than target column and label
        for j in range(start_meta_index, image_dataset.shape[1]):
            if has_dropped_features and column_names[j] in dropped_features:
                continue
            row_feature_values.append(image_dataset.iloc[i, j])
        results.append(row_feature_values)
    return results, feature_names
